#include "test_common.h"

CK_RV test_crypto();
CK_RV test_gost28147_pkcs8_key_wrap(CK_SESSION_HANDLE sess);

int main(int argc, char *argv[]) {
	CK_RV rc = CKR_OK;

	rc = test_main(argc, argv);
	if (rc != CKR_OK) {
		fprintf(stderr, "test_main failed: 0x%x\n", rc);
		return rc;
	}
	rc = test_crypto();
	if (rc != CKR_OK) {
		fprintf(stderr, "test_crypto failed: 0x%x\n", rc);
		return rc;
	}
	return CKR_OK;
}

CK_RV test_crypto() 
{
   CK_RV rc;
   CK_SESSION_HANDLE hSession;
	CK_MECHANISM_INFO minfo;

	printf(
"CKM_GOST28147_PKCS8_KEY_WRAP "
"and CKM_GOSTR3410_PUBLIC_KEY_DERIVE test\n");

	rc = funcs->C_GetMechanismInfo(SlotId, 
		CKM_GOSTR3410_KEY_PAIR_GEN, &minfo);
	if (rc != CKR_OK) {
		fprintf(stderr, 
			"Mechanism CKM_GOSTR3410_KEY_PAIR_GEN not supported\n");
		return rc;
	}
	rc = funcs->C_GetMechanismInfo(SlotId, 
		CKM_GOST28147_PKCS8_KEY_WRAP, &minfo);
	if (rc != CKR_OK) {
		fprintf(stderr, 
			"Mechanism CKM_GOST28147_PKCS8_KEY_WRAP not supported\n");
		return rc;
	}
	rc = funcs->C_GetMechanismInfo(SlotId, 
		CKM_GOSTR3410_PUBLIC_KEY_DERIVE, &minfo);
	if (rc != CKR_OK) {
		fprintf(stderr, 
			"Mechanism CKM_GOSTR3410_PUBLIC_KEY_DERIVE not supported\n");
		return rc;
	}

   rc = funcs->C_OpenSession(SlotId, 
	   CKF_RW_SESSION | CKF_SERIAL_SESSION, 
	   NULL_PTR, NULL_PTR, &hSession);
   if (rc != CKR_OK) {
		fprintf(stderr, 
			"ERROR call to C_OpenSession failed, rc = 0x%x\n", rc);
		goto out;
   }
/*
   rc = funcs->C_Login(hSession, CKU_USER, user_pin, strlen(user_pin));
   if (rc != CKR_OK) {
      fprintf(stderr, "ERROR call to C_Login failed, rc = 0x%x\n", rc);
      goto out_close;
   }
*/   
   rc = test_gost28147_pkcs8_key_wrap(hSession);

   if (rc != CKR_OK) {
      fprintf(stderr, 
"CKM_GOST28147_PKCS8_KEY_WRAP and CKM_GOSTR3410_PUBLIC_KEY_DERIVE "
"test failed, rc = 0x%x\n", 
		rc);
   } else {
      printf(
"CKM_GOST28147_PKCS8_KEY_WRAP and CKM_GOSTR3410_PUBLIC_KEY_DERIVE "
"test SUCCESS\n");
   }
 
   funcs->C_CloseSession(hSession);
out:
   return rc;
}

CK_RV test_gost28147_pkcs8_key_wrap(CK_SESSION_HANDLE sess)
{
	int rc = 0;
    char str[4096];
	CK_BYTE value[1024];
	CK_ULONG len;
	CK_OBJECT_HANDLE pub_key = CK_INVALID_HANDLE;
	CK_OBJECT_HANDLE priv_key = CK_INVALID_HANDLE;
	CK_OBJECT_HANDLE unw_priv_key = CK_INVALID_HANDLE;
    CK_OBJECT_HANDLE der_pub_key = CK_INVALID_HANDLE;
	CK_OBJECT_HANDLE cipher_key = CK_INVALID_HANDLE;
   	CK_MECHANISM mechanism_desc = 
		{CKM_GOST28147_PKCS8_KEY_WRAP, NULL, 0};
   	CK_MECHANISM_PTR mechanism = &mechanism_desc;
   	CK_MECHANISM mechanism_der_desc = 
		{CKM_GOSTR3410_PUBLIC_KEY_DERIVE, NULL, 0};
   	CK_MECHANISM_PTR mechanism_der = &mechanism_der_desc;
   	CK_MECHANISM mechanism_gen_256_desc = 
		{CKM_GOSTR3410_KEY_PAIR_GEN, NULL, 0};
   	CK_MECHANISM_PTR mechanism_gen_256 = &mechanism_gen_256_desc;
   	CK_MECHANISM mechanism_gen_512_desc = 
		{CKM_GOSTR3410_512_KEY_PAIR_GEN, NULL, 0};
   	CK_MECHANISM_PTR mechanism_gen_512 = &mechanism_gen_512_desc;

	static CK_BBOOL ltrue = CK_TRUE;
	static CK_BBOOL lfalse = CK_FALSE;
	static const CK_OBJECT_CLASS oclass_pub = CKO_PUBLIC_KEY;
	static const CK_OBJECT_CLASS oclass_priv = CKO_PRIVATE_KEY;
	static CK_KEY_TYPE key_type_256 = CKK_GOSTR3410;
	static CK_KEY_TYPE key_type_512 = CKK_GOSTR3410_512;
	// PAR 256 A OID
	static CK_BYTE gostR3410_256_params_A[] = { 
		0x06, 0x07, 0x2a, 0x85, 0x03, 0x02, 0x02, 0x23, 0x01 
	};
	// PAR 512 A OID
	CK_BYTE gostR3410_512_params_A[] = { 
		0x06, 0x09, 0x2a, 0x85, 0x03, 0x07, 0x01, 0x2, 0x01, 0x02, 0x01 
	};
	// PAR HASH 1 OID
	static CK_BYTE gostR3411_94_params[] = { 
		0x06, 0x07, 0x2a, 0x85, 0x03, 0x02, 0x02, 0x1e, 0x01 
	};
	// PAR 3411-2012-256 OID
	CK_BYTE gostR3411_2012_256_params[] = { 
		0x06, 0x08, 0x2a, 0x85, 0x03, 0x07, 0x01, 0x01, 0x02, 0x02
	};
	// PAR 3411-2012-512 OID
	CK_BYTE gostR3411_2012_512_params[] = { 
		0x06, 0x08, 0x2a, 0x85, 0x03, 0x07, 0x01, 0x01, 0x02, 0x03
	};
	// PAR CIPHER A OID
	static CK_BYTE gost28147params_A[] = { 
		0x06, 0x07, 0x2a, 0x85, 0x03, 0x02, 0x02, 0x1f, 0x01 
	};
/*
typedef struct CK_GOST28147_PKCS8_KEY_WRAP_PARAMS {
    CK_CHAR_PTR      pPassword;
    CK_ULONG         ulPasswordLen;
	CK_BYTE_PTR      pHashParOID;
	CK_ULONG         ulHashParOIDLen;
	CK_BYTE_PTR      pSalt;
	CK_ULONG         ulSaltLen;
    CK_ULONG         ulIterCount;
	CK_BYTE_PTR      pCipherParOID;
	CK_ULONG         ulCipherParOIDLen;
	CK_BYTE_PTR      pIV;
	CK_ULONG         ulIVLen;
} CK_GOST28147_PKCS8_KEY_WRAP_PARAMS;
*/
	CK_GOST28147_PKCS8_KEY_WRAP_PARAMS params;

	static CK_BYTE salt[] = { 
		0x12, 0x34, 0x56, 0x78, 0x78, 0x56, 0x34, 0x26
	};
	static CK_ULONG iter = 2048;
	static CK_BYTE iv[] = { 
		0x12, 0x34, 0x56, 0x78, 0x78, 0x56, 0x34, 0xab 
	};
    // Use UTF-8 password
	static CK_UTF8CHAR_PTR password = "4444";
	static CK_ULONG password_len = 4;
	CK_ATTRIBUTE pub_2001_256_template[] = {
		{ CKA_GOSTR3410PARAMS, 
		gostR3410_256_params_A, sizeof(gostR3410_256_params_A) },
		{ CKA_GOSTR3411PARAMS, 
		gostR3411_94_params, sizeof(gostR3411_94_params) },
	};
	CK_ATTRIBUTE pub_2012_256_template[] = {
		{ CKA_GOSTR3410PARAMS, 
		gostR3410_256_params_A, sizeof(gostR3410_256_params_A) },
		{ CKA_GOSTR3411PARAMS, 
		gostR3411_2012_256_params, sizeof(gostR3411_2012_256_params) },
	};
	CK_ATTRIBUTE pub_2012_512_template[] = {
		{ CKA_GOSTR3410PARAMS, 
		gostR3410_512_params_A, sizeof(gostR3410_512_params_A) },
		{ CKA_GOSTR3411PARAMS, 
		gostR3411_2012_512_params, sizeof(gostR3411_2012_512_params) },
	};
	// Template for token private key generation.
	CK_ATTRIBUTE priv_template[] = {
		{ CKA_EXTRACTABLE, &ltrue, sizeof(ltrue) },
		{ CKA_SENSITIVE, &ltrue, sizeof(ltrue) }
	};
	// Additional attributes for token unwrapped private key generation.
	CK_ATTRIBUTE unw_256_template[] = {
		{ CKA_CLASS, (CK_VOID_PTR)&oclass_priv, sizeof(oclass_priv) },
		{ CKA_KEY_TYPE, &key_type_256, sizeof(key_type_256) }
	};
	CK_ATTRIBUTE unw_512_template[] = {
		{ CKA_CLASS, (CK_VOID_PTR)&oclass_priv, sizeof(oclass_priv) },
		{ CKA_KEY_TYPE, &key_type_512, sizeof(key_type_512) }
	};
	CK_ATTRIBUTE pub_2001_256_der_template[] = {
        { CKA_CLASS, (CK_VOID_PTR)&oclass_pub, sizeof(oclass_pub) },
        { CKA_KEY_TYPE, &key_type_256, sizeof(key_type_256) },
		{ CKA_GOSTR3410PARAMS, 
		gostR3410_256_params_A, sizeof(gostR3410_256_params_A) },
		{ CKA_GOSTR3411PARAMS, 
		gostR3411_94_params, sizeof(gostR3411_94_params) },
	};
	CK_ATTRIBUTE pub_2012_256_der_template[] = {
        { CKA_CLASS, (CK_VOID_PTR)&oclass_pub, sizeof(oclass_pub) },
        { CKA_KEY_TYPE, &key_type_256, sizeof(key_type_256) },
		{ CKA_GOSTR3410PARAMS, 
		gostR3410_256_params_A, sizeof(gostR3410_256_params_A) },
		{ CKA_GOSTR3411PARAMS, 
		gostR3411_2012_256_params, sizeof(gostR3411_2012_256_params) },
	};
	CK_ATTRIBUTE pub_2012_512_der_template[] = {
        { CKA_CLASS, (CK_VOID_PTR)&oclass_pub, sizeof(oclass_pub) },
        { CKA_KEY_TYPE, &key_type_512, sizeof(key_type_512) },
		{ CKA_GOSTR3410PARAMS, 
		gostR3410_512_params_A, sizeof(gostR3410_512_params_A) },
		{ CKA_GOSTR3411PARAMS, 
		gostR3411_2012_512_params, sizeof(gostR3411_2012_512_params) },
	};
	// При упаковке на выходе должна получиться структура такого вида:
    static CK_BYTE pr_key_bag[] = {
// 0
        0x30, 0x81, 0xa7,
// 3
          0x30, 0x5c,
// 5
            0x06, 0x09, 
			  0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x05, 0x0d,
// 16
            0x30, 0x4f,
// 18
              0x30, 0x2e,
// 20
                0x06, 0x09, 
				  0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x05, 0x0c,
// 31
                0x30, 0x21,
// 33
                  0x04, 0x08,
// 35 - salt 
                    0x12, 0x34, 0x56, 0x78, 0x78, 0x56, 0x34, 0x26,
// 43
                  0x02, 0x02,
// 45 - iter (2048)
                    0x08, 0x00,
// 47
                  0x30, 0x11,
// 49
                    0x06, 0x06, 
					  0x2a, 0x85, 0x03, 0x02, 0x02, 0x0a,
// 57 - oid_3411_par 
                    0x06, 0x07, 
					  0x2a, 0x85, 0x03, 0x02, 0x02, 0x1e, 0x01,
// 66
                0x30, 0x1d,
// 68
                  0x06, 0x06, 
					0x2a, 0x85, 0x03, 0x02, 0x02, 0x15,
// 76
                  0x30, 0x13,
// 78
                    0x04, 0x08,
// 80 - iv
                      0x12, 0x34, 0x56, 0x78, 0x78, 0x56, 0x34, 0xab,
// 88 - oid_28147_par
                    0x06, 0x07, 
					  0x2a, 0x85, 0x03, 0x02, 0x02, 0x1f, 0x01,
// 97
          0x04, 0x47, 
// 99 - encr_pr_key
            0x46, 0x5c, 0xc8, 0x21, 0xcd, 0xcc, 0xb8, 0x99, 
            0x53, 0xa7, 0x72, 0xda, 0x20, 0x0f, 0x3d, 0xd5, 
            0xab, 0x59, 0x11, 0x6f, 0x4f, 0x8a, 0x75, 0x9b, 
            0xf4, 0xd1, 0x91, 0x7e, 0x9d, 0x2f, 0x79, 0xab, 
            0x95, 0xb8, 0x54, 0xe2, 0x5b, 0x21, 0x61, 0xa8, 
            0xe7, 0x2b, 0x58, 0x5b, 0xe2, 0x10, 0xd0, 0xd8, 
            0xbb, 0xd4, 0x04, 0xdc, 0x06, 0x4c, 0x60, 0x14, 
            0x60, 0x12, 0xd1, 0xf1, 0xb2, 0xbf, 0x39, 0xf7, 
            0xc8, 0x35, 0x16, 0xc4, 0xea, 0xe9, 0xb6
// 170
    };
	// Заметим, что при распаковке во входной структуре допускается 
	// использование 0x05, 0x00 вместо oid_3411_par. Соответственно,
	// длина такой укороченной структуры будет меньше на 7 байтов.

	// GOSTR3410-2001
	rc = funcs->C_GenerateKeyPair(sess, mechanism_gen_256,
			pub_2001_256_template, 
			sizeof(pub_2001_256_template)/sizeof(CK_ATTRIBUTE),
			priv_template, sizeof(priv_template)/sizeof(CK_ATTRIBUTE),
			&pub_key, &priv_key);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_GenerateKeyPair failed, rc = 0x%x\n", rc);
		goto end;
	}

	params.pPassword = password;
	params.ulPasswordLen = strlen(password);
	params.pHashParOID = gostR3411_94_params;
	params.ulHashParOIDLen = sizeof(gostR3411_94_params);
	params.pSalt = salt;
    params.ulSaltLen = sizeof(salt);
    params.ulIterCount = iter;
	params.pCipherParOID = gost28147params_A;
	params.ulCipherParOIDLen = sizeof(gost28147params_A);
    params.pIV = iv;
    params.ulIVLen = sizeof(iv);

	mechanism->pParameter = &params;
	mechanism->ulParameterLen = sizeof(params);
	rc = funcs->C_WrapKey(sess, mechanism, 
		CK_INVALID_HANDLE, priv_key, value, &len);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_WrapKey failed, rc = 0x%x\n", rc);
		goto end;
	}

    print_bytes(value, len, str);
    printf("PKCS#8:\n%s\n", str);
/*
	0x30, 0x81, 0xa7, 
		0x30, 0x5c, 
			0x06, 0x09, 
				0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x05, 0x0d,
			0x30, 0x4f, 
				0x30, 0x2e, 
					0x06, 0x09, 
						0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x05, 0x0c, 
					0x30, 0x21, 
						0x04, 0x08, 
							0x12, 0x34, 0x56, 0x78, 0x78, 0x56, 0x34, 0x26, 
						0x02, 0x02, 
							0x08, 0x00, 
						0x30, 0x11, 
							0x06, 0x06, 
								0x2a, 0x85, 0x03, 0x02, 0x02, 0x0a, 
							0x06, 0x07, 
								0x2a, 0x85, 0x03, 0x02, 0x02, 0x1e, 0x01, 
				0x30, 0x1d, 
					0x06, 0x06, 
						0x2a, 0x85, 0x03, 0x02, 0x02, 0x15, 
					0x30, 0x13, 
						0x04, 0x08,
							0x12, 0x34, 0x56, 0x78, 0x78, 0x56, 0x34, 0xab,
						0x06, 0x07, 
							0x2a, 0x85, 0x03, 0x02, 0x02, 0x1f, 0x01, 
		0x04, 0x47, 
			0x95, 0x44, 0x39, 0xe9, 0xce, 0xb5, 0x56, 0x29, 
			0xfa, 0x08, 0xf2, 0xab, 0x09, 0x81, 0x44, 0xca, 
			0xe0, 0x56, 0xc7, 0x60, 0xe2, 0x5c, 0xbd, 0x3c, 
			0xdd, 0x1c, 0x13, 0xf8, 0xf1, 0xe8, 0xd3, 0xf4, 
			0x98, 0xf0, 0x2e, 0x3c, 0xa0, 0x7d, 0xfe, 0x35, 
			0xee, 0xcf, 0xc4, 0xaf, 0xad, 0x93, 0x89, 0x48, 
			0xe5, 0xf2, 0x25, 0xbf, 0xe7, 0x61, 0x01, 0x8f, 
			0x50, 0x62, 0x08, 0x9b, 0xad, 0x9e, 0xcb, 0x42, 
			0xce, 0x17, 0x81, 0xc4, 0xea, 0xe9, 0xb6,
*/
    mechanism->pParameter = password;
    mechanism->ulParameterLen = strlen(password);
	rc = funcs->C_UnwrapKey(sess, mechanism, CK_INVALID_HANDLE, 
        value, len,
        unw_256_template, sizeof(unw_256_template)/sizeof(CK_ATTRIBUTE), 
        &unw_priv_key);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_UnwrapKey failed, rc = 0x%x\n", rc);
		goto end;
	}
	// Получаем открытый ключ по закрытому
	rc = funcs->C_DeriveKey(sess, mechanism_der, unw_priv_key, 
        pub_2001_256_der_template, 
		sizeof(pub_2001_256_der_template)/sizeof(CK_ATTRIBUTE), 
        &der_pub_key);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_DeriveKey failed, rc = 0x%x\n", rc);
		goto end;
	}
    // Для проверки распакованного ключа нужно что-нибудь им подписать
    // и проверить подпись открытым ключом.
    {
        CK_MECHANISM mechanism_sign = {CKM_GOSTR3410, NULL, 0};
		// Подписываемый дайджест (32 байта)
        CK_BYTE pData[32] = {
			0x4D, 0x89, 0x9E, 0x48, 0xC5, 0x39, 0x64, 0xD1, 
			0x78, 0xB4, 0x6D, 0x58, 0x40, 0x5F, 0x62, 0x8F, 
			0xA4, 0x46, 0x4C, 0xDC, 0x73, 0x75, 0xB0, 0xE5, 
			0x2E, 0x91, 0xFB, 0x64, 0x9C, 0x06, 0xAB, 0x75		
		};
        CK_ULONG ulDataLen = 32;
        CK_BYTE	pSignatureData[64];
        CK_ULONG signatureDataLen = 64;

        rc = funcs->C_SignInit(sess, &mechanism_sign, unw_priv_key);
		if (rc != CKR_OK) {
			fprintf(stderr, "C_SignInit failed, rc = 0x%x\n", rc);
			goto end;
		}
        rc = funcs->C_Sign(sess, 
			pData, ulDataLen, pSignatureData, &signatureDataLen);
		if (rc != CKR_OK) {
			fprintf(stderr, "C_Sign failed, rc = 0x%x\n", rc);
			goto end;
		}
	
        rc = funcs->C_VerifyInit(sess, &mechanism_sign, pub_key);
		if (rc != CKR_OK) {
			fprintf(stderr, "C_VerifyInit failed, rc = 0x%x\n", rc);
			goto end;
		}
        rc = funcs->C_Verify(sess, 
			pData, ulDataLen, pSignatureData, signatureDataLen);
		if (rc != CKR_OK) {
			fprintf(stderr, "C_Verify failed, rc = 0x%x\n", rc);
			goto end;
		}
 
        rc = funcs->C_VerifyInit(sess, &mechanism_sign, der_pub_key);
		if (rc != CKR_OK) {
			fprintf(stderr, "C_VerifyInit failed, rc = 0x%x\n", rc);
			goto end;
		}
        rc = funcs->C_Verify(sess, 
			pData, ulDataLen, pSignatureData, signatureDataLen);
		if (rc != CKR_OK) {
			fprintf(stderr, "C_Verify failed, rc = 0x%x\n", rc);
			goto end;
		}
    }
    funcs->C_DestroyObject(sess, der_pub_key);
    funcs->C_DestroyObject(sess, unw_priv_key);
	funcs->C_DestroyObject(sess, pub_key);
	funcs->C_DestroyObject(sess, priv_key);

	// GOSTR3410-2012-256
	rc = funcs->C_GenerateKeyPair(sess, mechanism_gen_256,
			pub_2012_256_template, 
			sizeof(pub_2012_256_template)/sizeof(CK_ATTRIBUTE),
			priv_template, sizeof(priv_template)/sizeof(CK_ATTRIBUTE),
			&pub_key, &priv_key);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_GenerateKeyPair failed, rc = 0x%x\n", rc);
		goto end;
	}

	params.pPassword = password;
	params.ulPasswordLen = strlen(password);
	params.pHashParOID = gostR3411_2012_256_params;
	params.ulHashParOIDLen = sizeof(gostR3411_2012_256_params);
	params.pSalt = salt;
    params.ulSaltLen = sizeof(salt);
    params.ulIterCount = iter;
	params.pCipherParOID = gost28147params_A;
	params.ulCipherParOIDLen = sizeof(gost28147params_A);
    params.pIV = iv;
    params.ulIVLen = sizeof(iv);

	mechanism->pParameter = &params;
	mechanism->ulParameterLen = sizeof(params);
	rc = funcs->C_WrapKey(sess, mechanism, 
		CK_INVALID_HANDLE, priv_key, value, &len);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_WrapKey failed, rc = 0x%x\n", rc);
		goto end;
	}

    print_bytes(value, len, str);
    printf("PKCS#8:\n%s\n", str);
/*
	0x30, 0x81, 0xab, 
		0x30, 0x5d, 
			0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x05, 0x0d,
			0x30, 0x50, 
				0x30, 0x2f, 
					0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x05, 0x0c, 
					0x30, 0x22, 
						0x04, 0x08, 
							0x12, 0x34, 0x56, 0x78, 0x78, 0x56, 0x34, 0x26, 
						0x02, 0x02, 
							0x08, 0x00, 
						0x30, 0x12, 
							0x06, 0x06, 0x2a, 0x85, 0x03, 0x02, 0x02, 0x0a, 
							0x06, 0x08, 0x2a, 0x85, 0x03, 0x07, 0x01, 0x01, 0x02, 0x02, 
				0x30, 0x1d, 
					0x06, 0x06, 0x2a, 0x85, 0x03, 0x02, 0x02, 0x15, 
					0x30, 0x13, 
						0x04, 0x08, 
							0x12, 0x34, 0x56, 0x78, 0x78, 0x56, 0x34, 0xab, 
						0x06, 0x07, 0x2a, 0x85, 0x03, 0x02, 0x02, 0x1f, 0x01, 
		0x04, 0x4a, 
			0x68, 0x7f, 0x0d, 0x40, 0xbc, 0x97, 0xe6, 0x83, 
			0x83, 0xdf, 0x4b, 0x9e, 0xff, 0x60, 0x64, 0xbd, 
			0xb3, 0xea, 0xfe, 0x6b, 0xb2, 0x00, 0xc7, 0x78, 
			0x80, 0x13, 0xc7, 0x61, 0x45, 0x76, 0x30, 0x53, 
			0x5d, 0xf1, 0xcf, 0xdc, 0xf3, 0xd9, 0xdd, 0x50, 
			0x7d, 0x7a, 0x2c, 0x87, 0x06, 0x2f, 0x28, 0xdb, 
			0x1e, 0x94, 0xe3, 0xab, 0x9f, 0x0a, 0x75, 0x09, 
			0xd3, 0xf9, 0x2e, 0xbe, 0xc6, 0xb1, 0xf4, 0x9c, 
			0x92, 0xc9, 0xa4, 0x35, 0xd6, 0x6c, 0x25, 0xc7, 
			0xbc, 0xcf,
*/
    mechanism->pParameter = password;
    mechanism->ulParameterLen = strlen(password);
	rc = funcs->C_UnwrapKey(sess, mechanism, CK_INVALID_HANDLE, 
        value, len,
        unw_256_template, 
		sizeof(unw_256_template)/sizeof(CK_ATTRIBUTE), 
        &unw_priv_key);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_UnwrapKey failed, rc = 0x%x\n", rc);
		goto end;
	}
	// Получаем открытый ключ по закрытому
	rc = funcs->C_DeriveKey(sess, mechanism_der, unw_priv_key, 
        pub_2012_256_der_template, 
		sizeof(pub_2012_256_der_template)/sizeof(CK_ATTRIBUTE), 
        &der_pub_key);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_DeriveKey failed, rc = 0x%x\n", rc);
		goto end;
	}
    // Для проверки распакованного ключа нужно что-нибудь им подписать
    // и проверить подпись открытым ключом.
    {
        CK_MECHANISM mechanism_sign = {CKM_GOSTR3410, NULL, 0};
		// Подписываемый дайджест (32 байта)
        CK_BYTE pData[32] = {
			0x4D, 0x89, 0x9E, 0x48, 0xC5, 0x39, 0x64, 0xD1, 
			0x78, 0xB4, 0x6D, 0x58, 0x40, 0x5F, 0x62, 0x8F, 
			0xA4, 0x46, 0x4C, 0xDC, 0x73, 0x75, 0xB0, 0xE5, 
			0x2E, 0x91, 0xFB, 0x64, 0x9C, 0x06, 0xAB, 0x75		
		};
        CK_ULONG ulDataLen = 32;
        CK_BYTE	pSignatureData[64];
        CK_ULONG signatureDataLen = 64;

        rc = funcs->C_SignInit(sess, &mechanism_sign, unw_priv_key);
		if (rc != CKR_OK) {
			fprintf(stderr, "C_SignInit failed, rc = 0x%x\n", rc);
			goto end;
		}
        rc = funcs->C_Sign(sess, 
			pData, ulDataLen, pSignatureData, &signatureDataLen);
		if (rc != CKR_OK) {
			fprintf(stderr, "C_Sign failed, rc = 0x%x\n", rc);
			goto end;
		}
	
        rc = funcs->C_VerifyInit(sess, &mechanism_sign, pub_key);
		if (rc != CKR_OK) {
			fprintf(stderr, "C_VerifyInit failed, rc = 0x%x\n", rc);
			goto end;
		}
        rc = funcs->C_Verify(sess, 
			pData, ulDataLen, pSignatureData, signatureDataLen);
		if (rc != CKR_OK) {
			fprintf(stderr, "C_Verify failed, rc = 0x%x\n", rc);
			goto end;
		}
 
        rc = funcs->C_VerifyInit(sess, &mechanism_sign, der_pub_key);
		if (rc != CKR_OK) {
			fprintf(stderr, "C_VerifyInit failed, rc = 0x%x\n", rc);
			goto end;
		}
        rc = funcs->C_Verify(sess, 
			pData, ulDataLen, pSignatureData, signatureDataLen);
		if (rc != CKR_OK) {
			fprintf(stderr, "C_Verify failed, rc = 0x%x\n", rc);
			goto end;
		}
    }
    funcs->C_DestroyObject(sess, der_pub_key);
    funcs->C_DestroyObject(sess, unw_priv_key);
	funcs->C_DestroyObject(sess, pub_key);
	funcs->C_DestroyObject(sess, priv_key);

	// GOSTR3410-2012-512
	rc = funcs->C_GenerateKeyPair(sess, mechanism_gen_512,
			pub_2012_512_template, 
			sizeof(pub_2012_512_template)/sizeof(CK_ATTRIBUTE),
			priv_template, sizeof(priv_template)/sizeof(CK_ATTRIBUTE),
			&pub_key, &priv_key);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_GenerateKeyPair failed, rc = 0x%x\n", rc);
		goto end;
	}

	params.pPassword = password;
	params.ulPasswordLen = strlen(password);
	params.pHashParOID = gostR3411_2012_512_params;
	params.ulHashParOIDLen = sizeof(gostR3411_2012_512_params);
	params.pSalt = salt;
    params.ulSaltLen = sizeof(salt);
    params.ulIterCount = iter;
	params.pCipherParOID = gost28147params_A;
	params.ulCipherParOIDLen = sizeof(gost28147params_A);
    params.pIV = iv;
    params.ulIVLen = sizeof(iv);

	mechanism->pParameter = &params;
	mechanism->ulParameterLen = sizeof(params);
	rc = funcs->C_WrapKey(sess, mechanism, 
		CK_INVALID_HANDLE, priv_key, value, &len);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_WrapKey failed, rc = 0x%x\n", rc);
		goto end;
	}

    print_bytes(value, len, str);
    printf("PKCS#8:\n%s\n", str);
/*
	0x30, 0x81, 0xcd, 
		0x30, 0x5d, 
			0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x05, 0x0d,
			0x30, 0x50, 
				0x30, 0x2f, 
					0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x05, 0x0c, 
					0x30, 0x22, 
						0x04, 0x08, 
							0x12, 0x34, 0x56, 0x78, 0x78, 0x56, 0x34, 0x26, 
						0x02, 0x02, 
							0x08, 0x00, 
				0x30, 0x12, 
					0x06, 0x06, 0x2a, 0x85, 0x03, 0x02, 0x02, 0x0a, 
					0x06, 0x08, 0x2a, 0x85, 0x03, 0x07, 0x01, 0x01, 0x02, 0x03, 
			0x30, 0x1d, 
				0x06, 0x06, 0x2a, 0x85, 0x03, 0x02, 0x02, 0x15, 
				0x30, 0x13, 
					0x04, 0x08, 
						0x12, 0x34, 0x56, 0x78, 0x78, 0x56, 0x34, 0xab, 
					0x06, 0x07, 0x2a, 0x85, 0x03, 0x02, 0x02, 0x1f, 0x01, 
		0x04, 0x6c, 
			0x68, 0x5d, 0x0d, 0x40, 0xbc, 0x97, 0xd8, 0x83, 
			0x73, 0x78, 0x9a, 0x75, 0x9a, 0xa1, 0x44, 0xf1, 
			0x80, 0x83, 0x05, 0x32, 0x1a, 0x83, 0x36, 0x03, 
			0x9e, 0x76, 0x41, 0xb8, 0x56, 0x10, 0x25, 0x6c, 
			0x65, 0x83, 0xa8, 0x46, 0xd5, 0x1b, 0xb0, 0x04, 
			0x41, 0x5c, 0x13, 0x91, 0x76, 0x1c, 0x85, 0x59, 
			0x6b, 0x1d, 0x8d, 0xc8, 0xde, 0x93, 0x2e, 0x22, 
			0x1c, 0x5c, 0xec, 0xaf, 0x1f, 0xb7, 0x7b, 0x63, 
			0xa5, 0x72, 0x1f, 0x29, 0xd6, 0x07, 0xf6, 0xdf, 
			0x8d, 0x69, 0xe4, 0x49, 0x35, 0xfb, 0x71, 0xda, 
			0xb7, 0x4b, 0x17, 0x8c, 0xae, 0xbd, 0xcf, 0x62, 
			0xd2, 0xe6, 0xdb, 0x4d, 0x7c, 0x47, 0x46, 0x82, 
			0xdb, 0x0c, 0xcd, 0x5b, 0x0a, 0xc0, 0xd8, 0x42, 
			0x58, 0x77, 0xe7, 0x7e,
*/
    mechanism->pParameter = password;
    mechanism->ulParameterLen = strlen(password);
	rc = funcs->C_UnwrapKey(sess, mechanism, CK_INVALID_HANDLE, 
        value, len,
        unw_512_template, sizeof(unw_512_template)/sizeof(CK_ATTRIBUTE), 
        &unw_priv_key);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_UnwrapKey failed, rc = 0x%x\n", rc);
		goto end;
	}
	// Получаем открытый ключ по закрытому
	rc = funcs->C_DeriveKey(sess, mechanism_der, unw_priv_key, 
        pub_2012_512_der_template, 
		sizeof(pub_2012_512_der_template)/sizeof(CK_ATTRIBUTE), 
        &der_pub_key);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_DeriveKey failed, rc = 0x%x\n", rc);
		goto end;
	}
    // Для проверки распакованного ключа нужно что-нибудь им подписать
    // и проверить подпись открытым ключом.
    {
        CK_MECHANISM mechanism_sign = {CKM_GOSTR3410_512, NULL, 0};
		// Подписываемый дайджест (64 байта)
        CK_BYTE pData[] = {
			0x4D, 0x89, 0x9E, 0x48, 0xC5, 0x39, 0x64, 0xD1, 
			0x78, 0xB4, 0x6D, 0x58, 0x40, 0x5F, 0x62, 0x8F, 
			0xA4, 0x46, 0x4C, 0xDC, 0x73, 0x75, 0xB0, 0xE5, 
			0x2E, 0x91, 0xFB, 0x64, 0x9C, 0x06, 0xAB, 0x75,		
			0x4D, 0x89, 0x9E, 0x48, 0xC5, 0x39, 0x64, 0xD1, 
			0x78, 0xB4, 0x6D, 0x58, 0x40, 0x5F, 0x62, 0x8F, 
			0xA4, 0x46, 0x4C, 0xDC, 0x73, 0x75, 0xB0, 0xE5, 
			0x2E, 0x91, 0xFB, 0x64, 0x9C, 0x06, 0xAB, 0x75,		
		};
        CK_ULONG ulDataLen = 64;
        CK_BYTE	pSignatureData[128];
        CK_ULONG signatureDataLen = 128;

        rc = funcs->C_SignInit(sess, &mechanism_sign, unw_priv_key);
		if (rc != CKR_OK) {
			fprintf(stderr, "C_SignInit failed, rc = 0x%x\n", rc);
			goto end;
		}
        rc = funcs->C_Sign(sess, 
			pData, ulDataLen, pSignatureData, &signatureDataLen);
		if (rc != CKR_OK) {
			fprintf(stderr, "C_Sign failed, rc = 0x%x\n", rc);
			goto end;
		}
	
        rc = funcs->C_VerifyInit(sess, &mechanism_sign, pub_key);
		if (rc != CKR_OK) {
			fprintf(stderr, "C_VerifyInit failed, rc = 0x%x\n", rc);
			goto end;
		}
        rc = funcs->C_Verify(sess, 
			pData, ulDataLen, pSignatureData, signatureDataLen);
		if (rc != CKR_OK) {
			fprintf(stderr, "C_Verify failed, rc = 0x%x\n", rc);
			goto end;
		}
 
        rc = funcs->C_VerifyInit(sess, &mechanism_sign, der_pub_key);
		if (rc != CKR_OK) {
			fprintf(stderr, "C_VerifyInit failed, rc = 0x%x\n", rc);
			goto end;
		}
        rc = funcs->C_Verify(sess, 
			pData, ulDataLen, pSignatureData, signatureDataLen);
		if (rc != CKR_OK) {
			fprintf(stderr, "C_Verify failed, rc = 0x%x\n", rc);
			goto end;
		}
    }
    funcs->C_DestroyObject(sess, der_pub_key);
    funcs->C_DestroyObject(sess, unw_priv_key);
	funcs->C_DestroyObject(sess, pub_key);
	funcs->C_DestroyObject(sess, priv_key);

	printf("SUCCESS\n");
end:

    return rc;
}
