#include "test_common.h"

CK_RV test_crypto();
CK_RV test_gost3411_12_256_hmac(CK_SESSION_HANDLE sess);

int main(int argc, char *argv[]) {
	CK_RV rc = CKR_OK;

	rc = test_main(argc, argv);
	if (rc != CKR_OK) {
		fprintf(stderr, "test_main failed: 0x%x\n", rc);
		return rc;
	}
	rc = test_crypto();
	if (rc != CKR_OK) {
		fprintf(stderr, "test_crypto failed: 0x%x\n", rc);
		return rc;
	}
	return CKR_OK;
}

CK_RV test_crypto() 
{
	CK_RV rc, ret;
	CK_SESSION_HANDLE hSession;
	CK_MECHANISM_INFO minfo;

	rc = funcs->C_OpenSession(SlotId, 
		CKF_RW_SESSION | CKF_SERIAL_SESSION, 
		NULL_PTR, NULL_PTR, &hSession);
	if (rc != CKR_OK) {
		fprintf(stderr, 
			"ERROR call to C_OpenSession failed, rc = 0x%x\n", rc);
		goto out;
	}
	fprintf(stderr, "C_OpenSession success\n");
 
	rc = funcs->C_GetMechanismInfo(SlotId, 
		CKM_GOSTR3411_12_256_HMAC, &minfo);
	if (rc != CKR_OK) {
		fprintf(stderr, 
"\n======== Mechanism CKM_GOSTR3411_12_256_HMAC "
"not supported ===========\n");
	} else {
		fprintf(stderr, 
"\n==================== CKM_GOSTR3411_12_256_HMAC "
"test ==================\n");
		rc = test_gost3411_12_256_hmac(hSession);
		if (rc != CKR_OK) {
			fprintf(stderr, 
				"ERROR CKM_GOSTR3411_12_256_HMAC failed, rc = 0x%x\n",
				rc);
		} else {
			fprintf(stderr, 
				"CKM_GOSTR3411_12_256_HMAC test passed.\n");
		}
	}
 
	if( (ret = funcs->C_CloseSession(hSession)) != CKR_OK ) {
		fprintf(stderr, 
			"Error: C_CloseSession failed with 0x%x\n", ret);
		rc = ret;
	}
	else {
		fprintf(stderr, "C_CloseSession success\n");
	}

out:
	return rc;
}

CK_RV test_gost3411_12_256_hmac(CK_SESSION_HANDLE sess)
{
	int rc = 0;
	CK_BYTE value[256];
	CK_ULONG len;
   	CK_MECHANISM mechanism_desc = 
		{CKM_GOSTR3411_12_256_HMAC, NULL, 0};
   	CK_MECHANISM_PTR mechanism = &mechanism_desc;

	CK_OBJECT_HANDLE keyh = CK_INVALID_HANDLE;
	static CK_OBJECT_CLASS oclass = CKO_SECRET_KEY;
	static CK_KEY_TYPE key_type = CKK_GENERIC_SECRET;
	static CK_BBOOL ltrue = CK_TRUE;

	static CK_BYTE keyval1[] = { 
		0x0b,0x0b,0x0b,0x0b,0x0b,0x0b,0x0b,0x0b, 
		0x0b,0x0b,0x0b,0x0b,0x0b,0x0b,0x0b,0x0b, 
		0x0b,0x0b,0x0b,0x0b,
	};
	static CK_BYTE data1[] = "Hi There";
	static CK_BYTE et1[] = {
		0x79, 0x8b, 0xd0, 0x9e, 0xf7, 0xad, 0x67, 0x64,
		0x6a, 0x3b, 0x74, 0xfe, 0x48, 0xd7, 0x31, 0x73,
		0x50, 0xa8, 0x6f, 0x07, 0x7d, 0x05, 0xef, 0x78,
		0xcb, 0xe9, 0xba, 0x3c, 0x9e, 0xf3, 0xa4, 0x11,
	};

	static CK_BYTE keyval2[] = { 'J','e','f','e' };
	static CK_BYTE data2[] = "what do ya want for nothing?";
	static CK_BYTE et2[] = {
		0x00,0x61,0x75,0x04,0x8e,0x45,0x72,0xac,
		0x61,0x85,0xfa,0xf5,0xff,0xae,0x49,0x62,
		0x97,0xf9,0x99,0x3f,0xf1,0xab,0xb7,0x05,
		0xce,0x43,0xda,0x09,0x51,0x56,0x8d,0x9a,
	};

	static CK_BYTE keyval3[] = {
		0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,
		0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,
		0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,
		0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,
		0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,
		0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,
		0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,
		0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,
		0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,
		0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,
	};
	static CK_BYTE data3[] = 
		"Test Using Larger Than Block-Size Key - Hash Key First";
	static CK_BYTE et3[] = {
		0x91,0x77,0xa0,0x5f,0xee,0xca,0xfa,0x5c,
		0xbb,0x14,0x8b,0x17,0x61,0x63,0x70,0xb5,
		0xd5,0x9b,0x23,0x4c,0x56,0x7e,0x26,0x0d,
		0xda,0xe2,0x67,0x2a,0x5f,0xd3,0x45,0xfa,
	};
	static CK_ATTRIBUTE key_template[] = {
		{ CKA_VALUE, NULL_PTR, 0 },
		{ CKA_CLASS, &oclass, sizeof(oclass) },
		{ CKA_KEY_TYPE, &key_type, sizeof(key_type) },
		{ CKA_SIGN, &ltrue, sizeof(ltrue) },
		{ CKA_VERIFY, &ltrue, sizeof(ltrue) }
	};
	
	static CK_BYTE data4[] = {
	    0x61, 0x62, 0x63, 0x64, 0x62, 0x63, 0x64, 0x65,
	    0x63, 0x64, 0x65, 0x66, 0x64, 0x65, 0x66, 0x67,
	    0x65, 0x66, 0x67, 0x68, 0x66, 0x67, 0x68, 0x69,
	    0x67, 0x68, 0x69, 0x6A, 0x68, 0x69, 0x6A, 0x6B,
	    0x69, 0x6A, 0x6B, 0x6C, 0x6A, 0x6B, 0x6C, 0x6D,
	    0x6B, 0x6C, 0x6D, 0x6E, 0x6C, 0x6D, 0x6E, 0x6F,
	    0x6D, 0x6E, 0x6F, 0x70, 0x6E, 0x6F, 0x70, 0x71,
	    0x0A
	};
	static CK_BYTE keyval_32[] = {
	    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
	    0x38, 0x39, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66,
	    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
	    0x38, 0x39, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66
	};
	// For big-endian byte order
	static CK_BYTE et_32[] = {
	    0x52, 0x67, 0xfb, 0xa5, 0xbf, 0x8c, 0x73, 0xaf,
	    0x26, 0x2c, 0xc2, 0xa0, 0x45, 0x61, 0x86, 0x2e,
	    0x28, 0xce, 0xd8, 0xe9, 0x6f, 0x0a, 0x85, 0xf8,
	    0xf8, 0x8b, 0x59, 0x40, 0xb8, 0xf5, 0x4a, 0x56
	};
	// For little-endian byte order
	static CK_BYTE et_32_le[] = {
		0xab, 0x42, 0x71, 0xd2, 0x71, 0xd0, 0xbb, 0x78,
		0x67, 0x5e, 0xe8, 0x10, 0x33, 0x7e, 0x5c, 0xd3,
		0xcf, 0x1c, 0x3c, 0x2b, 0x7d, 0x12, 0xd1, 0xdd,
		0xf1, 0xec, 0x31, 0xf9, 0xfc, 0x00, 0x5b, 0xe2,
	};
	static CK_BYTE keyval_31[] = {
	    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
	    0x38, 0x39, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66,
	    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
	    0x38, 0x39, 0x61, 0x62, 0x63, 0x64, 0x65
	};
	// For big-endian byte order
	static CK_BYTE et_31[] = {
	    0x98, 0x24, 0x2e, 0xd7, 0x01, 0x14, 0xc6, 0x94,
	    0xbf, 0xfd, 0x10, 0x7b, 0x2f, 0x7e, 0xc3, 0xc7,
	    0x87, 0x14, 0x61, 0xaf, 0x7f, 0x39, 0xf4, 0x05,
	    0xe7, 0x2b, 0xfb, 0x63, 0x36, 0x15, 0xf1, 0xce
	};
	// For little-endian byte order
	static CK_BYTE et_31_le[] = {
		0x04, 0x9a, 0xe0, 0xf4, 0xa2, 0xba, 0xac, 0x75,
		0x82, 0xd5, 0x5d, 0x46, 0x45, 0x47, 0x17, 0x71,
		0xec, 0x77, 0xbc, 0x24, 0xa2, 0xb8, 0x5c, 0xbd,
		0xaa, 0x53, 0x0c, 0xb6, 0x85, 0x50, 0x4c, 0x9c,
	};
	static CK_BYTE keyval_33[] = {
	    0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38,
	    0x39, 0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36,
	    0x37, 0x38, 0x39, 0x30, 0x31, 0x32, 0x33, 0x34,
	    0x35, 0x36, 0x37, 0x38, 0x39, 0x30, 0x31, 0x32,
	    0x0A
	};
	static CK_BYTE keyval_33_hash[] = {
		0x72, 0xc3, 0x40, 0xc5, 0x8f, 0xbb, 0x9d, 0x91,
		0xdf, 0x95, 0x7c, 0xff, 0x0a, 0x87, 0xaa, 0x45,
		0xde, 0xd3, 0xb9, 0x78, 0x12, 0xe8, 0x41, 0xc9,
		0xd4, 0x97, 0x7f, 0xd0, 0xe2, 0xd8, 0xea, 0xb4,
	};
	// For big-endian byte order
	static CK_BYTE et_33[] = {
	    0x18, 0x8f, 0x6c, 0xc3, 0x3f, 0x19, 0xbe, 0x1b,
	    0x00, 0xa0, 0x67, 0x80, 0x90, 0xdb, 0xb2, 0x2b,
	    0x5a, 0xaa, 0xb3, 0xec, 0x4e, 0x97, 0x5f, 0x6a,
	    0xa5, 0xe5, 0x2b, 0xf3, 0x39, 0x57, 0x72, 0xeb
	};
	// For little-endian byte order
	static CK_BYTE et_33_le[] = {
		0xbb, 0x9c, 0x11, 0x4a, 0x86, 0x5e, 0x3f, 0x1b,
		0xec, 0x2a, 0xd9, 0x68, 0x3a, 0x11, 0x14, 0xaf,
		0x8f, 0x6d, 0xed, 0x1a, 0x71, 0x4e, 0xa5, 0x41,
		0x01, 0x71, 0x3c, 0x9d, 0xc7, 0x89, 0x32, 0x1b,	
	};
	CK_BYTE *state = NULL;
	CK_ULONG state_len = 0;

	key_template[0].pValue = keyval_32; 
	key_template[0].ulValueLen = sizeof(keyval_32);
	rc = funcs->C_CreateObject(sess, 
		key_template, sizeof(key_template)/sizeof(CK_ATTRIBUTE), 
		&keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_CreateObject failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_SignInit(sess, mechanism, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_SignInit failed: 0x%x\n", rc);
		goto end;
	}

	memset(value,0,sizeof(value));
	len = sizeof(value);
	rc = funcs->C_Sign(sess, data4, sizeof(data4), 
		value, &len);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_Sign failed: 0x%x\n", rc);
		goto end;
	}

	if (len != sizeof(et_32_le)) {
		fprintf(stderr, "Invalid length: %d\n", len);
		rc = -2;
		goto end;
	}
	print_hex(value, len);
	if (memcmp(value, et_32_le, len) != 0) {
		fprintf(stderr, "Invalid value\n");
		rc = -3;
		goto end;
	}
	printf("Sign 32_le OK\n");

	rc = funcs->C_VerifyInit(sess, mechanism, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_VerifyInit failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_Verify(sess, data4, sizeof(data4), 
		value, len);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_Verify failed: 0x%x\n", rc);
		goto end;
	}
	printf("Verify 32_le OK\n");

	rc = funcs->C_DestroyObject(sess, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_DestroyObject failed: 0x%x\n", rc);
		goto end;
	}

	key_template[0].pValue = keyval_31; 
	key_template[0].ulValueLen = sizeof(keyval_31);
	rc = funcs->C_CreateObject(sess, 
		key_template, sizeof(key_template)/sizeof(CK_ATTRIBUTE), 
		&keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_CreateObject failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_SignInit(sess, mechanism, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_SignInit failed: 0x%x\n", rc);
		goto end;
	}

	memset(value,0,sizeof(value));
	len = sizeof(value);
	rc = funcs->C_Sign(sess, data4, sizeof(data4), 
		value, &len);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_Sign failed: 0x%x\n", rc);
		goto end;
	}

	if (len != sizeof(et_31_le)) {
		fprintf(stderr, "Invalid length: %d\n", len);
		rc = -4;
		goto end;
	}
	print_hex(value, len);
	if (memcmp(value, et_31_le, len) != 0) {
		fprintf(stderr, "Invalid value\n");
		rc = -5;
		goto end;
	}
	printf("Sign 31_le OK\n");

	rc = funcs->C_VerifyInit(sess, mechanism, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_VerifyInit failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_Verify(sess, data4, sizeof(data4), 
		value, len);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_Verify failed: 0x%x\n", rc);
		goto end;
	}
	printf("Verify 31_le OK\n");

	rc = funcs->C_DestroyObject(sess, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_DestroyObject failed: 0x%x\n", rc);
		goto end;
	}

	key_template[0].pValue = keyval_33; 
	key_template[0].ulValueLen = sizeof(keyval_33);
	rc = funcs->C_CreateObject(sess, 
		key_template, sizeof(key_template)/sizeof(CK_ATTRIBUTE), 
		&keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_CreateObject failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_SignInit(sess, mechanism, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_SignInit failed: 0x%x\n", rc);
		goto end;
	}

	memset(value,0,sizeof(value));
	len = sizeof(value);
	rc = funcs->C_Sign(sess, data4, sizeof(data4), value, &len);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_Sign failed: 0x%x\n", rc);
		goto end;
	}

	if (len != sizeof(et_33_le)) {
		fprintf(stderr, "Invalid length: %d\n", len);
		rc = -8;
		goto end;
	}
	print_hex(value, len);
	if (memcmp(value, et_33_le, len) != 0) {
		fprintf(stderr, "Invalid value\n");
		rc = -9;
		goto end;
	}

	rc = funcs->C_VerifyInit(sess, mechanism, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_VerifyInit failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_Verify(sess, data4, sizeof(data4), value, len);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_Verify failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_DestroyObject(sess, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_DestroyObject failed: 0x%x\n", rc);
		goto end;
	}

	printf("keyval_33_hash multipart test\n");
	key_template[0].pValue = keyval_33_hash; 
	key_template[0].ulValueLen = sizeof(keyval_33_hash);
	rc = funcs->C_CreateObject(sess, 
		key_template, sizeof(key_template)/sizeof(CK_ATTRIBUTE), 
		&keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_CreateObject failed: 0x%x\n", rc);
		goto end;
	}

	mechanism->pParameter = 0;
	mechanism->ulParameterLen = 0;
	rc = funcs->C_SignInit(sess, mechanism, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_SignInit failed: 0x%x\n", rc);
		goto end;
	}

	memset(value,0,sizeof(value));
	len = sizeof(value);
	rc = funcs->C_SignUpdate(sess, data4, 7);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_SignUpdate failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_GetOperationState(sess, NULL, &state_len);
	if (rc != CKR_OK) {
		fprintf(stderr, 
			"C_GetOperationState failed: 0x%x\n", rc);
		goto end;
	}
	state = (CK_BYTE *)malloc(state_len);
	rc = funcs->C_GetOperationState(sess, state, &state_len);
	if (rc != CKR_OK) {
		fprintf(stderr, 
			"C_GetOperationState failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_SetOperationState(sess, state, state_len,
		CK_INVALID_HANDLE, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, 
			"C_SetOperationState failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_SignUpdate(sess, data4+7, sizeof(data4)-7);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_SignUpdate failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_SignFinal(sess, value, &len);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_SignFinal failed: 0x%x\n", rc);
		goto end;
	}

	if (len != sizeof(et_33_le)) {
		fprintf(stderr, "Invalid length: %d\n", len);
		rc = -6;
		goto end;
	}
	print_hex(value, len);
	if (memcmp(value, et_33_le, len) != 0) {
		fprintf(stderr, "Invalid value\n");
		rc = -7;
		goto end;
	}
	printf("Sign 33_le multipart OK\n");

	mechanism->pParameter = 0;
	mechanism->ulParameterLen = 0;
	rc = funcs->C_SignInit(sess, mechanism, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_SignInit failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_SetOperationState(sess, state, state_len,
		CK_INVALID_HANDLE, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_SetOperationState failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_SignUpdate(sess, data4+7, sizeof(data4)-7);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_SignUpdate failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_SignFinal(sess, value, &len);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_SignFinal failed: 0x%x\n", rc);
		goto end;
	}

	if (len != sizeof(et_33_le)) {
		fprintf(stderr, "Invalid length: %d\n", len);
		rc = -6;
		goto end;
	}
	print_hex(value, len);
	if (memcmp(value, et_33_le, len) != 0) {
		fprintf(stderr, "Invalid value\n");
		rc = -7;
		goto end;
	}
	printf("Restored Sign 33_le multipart OK\n");

	rc = funcs->C_VerifyInit(sess, mechanism, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_VerifyInit failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_Verify(sess, data4, sizeof(data4), value, len);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_Verify failed: 0x%x\n", rc);
		goto end;
	}
	printf("Verify 33_le OK\n");

	rc = funcs->C_DestroyObject(sess, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_DestroyObject failed: 0x%x\n", rc);
		goto end;
	}

	printf("keyval_33 singlepart test\n");
	key_template[0].pValue = keyval_33; 
	key_template[0].ulValueLen = sizeof(keyval_33);
	rc = funcs->C_CreateObject(sess, 
		key_template, sizeof(key_template)/sizeof(CK_ATTRIBUTE), 
		&keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_CreateObject failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_SignInit(sess, mechanism, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_SignInit failed: 0x%x\n", rc);
		goto end;
	}

	memset(value,0,sizeof(value));
	len = sizeof(value);
	rc = funcs->C_Sign(sess, data4, sizeof(data4), value, &len);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_Sign failed: 0x%x\n", rc);
		goto end;
	}

	if (len != sizeof(et_33_le)) {
		fprintf(stderr, "Invalid length: %d\n", len);
		rc = -8;
		goto end;
	}
	print_hex(value, len);
	if (memcmp(value, et_33_le, len) != 0) {
		fprintf(stderr, "Invalid value\n");
		rc = -9;
		goto end;
	}
	printf("keyval_33 singlepart sign OK\n");

	rc = funcs->C_VerifyInit(sess, mechanism, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_VerifyInit failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_Verify(sess, data4, sizeof(data4), value, len);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_Verify failed: 0x%x\n", rc);
		goto end;
	}
	printf("keyval_33 singlepart verify OK\n");

	rc = funcs->C_DestroyObject(sess, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_DestroyObject failed: 0x%x\n", rc);
		goto end;
	}

	printf("keyval1 singlepart sign test\n");
	key_template[0].pValue = keyval1; 
	key_template[0].ulValueLen = sizeof(keyval1);
	rc = funcs->C_CreateObject(sess, 
		key_template, sizeof(key_template)/sizeof(CK_ATTRIBUTE),
		&keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_CreateObject failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_SignInit(sess, mechanism, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_SignInit failed: 0x%x\n", rc);
		goto end;
	}

	memset(value,0,sizeof(value));
	len = sizeof(value);
	rc = funcs->C_Sign(sess, data1, strlen(data1), value, &len);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_Sign failed: 0x%x\n", rc);
		goto end;
	}

	if (len != sizeof(et1)) {
		fprintf(stderr, "Invalid length: %d\n", len);
		rc = -10;
		goto end;
	}
	print_hex(value, len);
	if (memcmp(value, et1, len) != 0) {
		fprintf(stderr, "Invalid value\n");
		rc = -11;
		goto end;
	}
	printf("keyval1 singlepart sign OK\n");

	rc = funcs->C_VerifyInit(sess, mechanism, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_VerifyInit failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_Verify(sess, data1, strlen(data1), value, len);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_Verify failed: 0x%x\n", rc);
		goto end;
	}
	printf("keyval1 singlepart verify OK\n");

	printf("SUCCESS\n");
	rc = CKR_OK;
end:
	if (keyh != CK_INVALID_HANDLE) {
		funcs->C_DestroyObject(sess, keyh);
	}

	return rc;
}

