#include "test_common.h"

CK_RV test_crypto();
CK_RV test_gost3411_12_512_hmac(CK_SESSION_HANDLE sess);

int main(int argc, char *argv[]) {
	CK_RV rc = CKR_OK;

	rc = test_main(argc, argv);
	if (rc != CKR_OK) {
		fprintf(stderr, "test_main failed: 0x%x\n", rc);
		return rc;
	}
	rc = test_crypto();
	if (rc != CKR_OK) {
		fprintf(stderr, "test_crypto failed: 0x%x\n", rc);
		return rc;
	}
	return CKR_OK;
}

CK_RV test_crypto() 
{
	CK_RV rc, ret;
	CK_SESSION_HANDLE hSession;
	CK_MECHANISM_INFO minfo;

	rc = funcs->C_OpenSession(SlotId, 
		CKF_RW_SESSION | CKF_SERIAL_SESSION, 
		NULL_PTR, NULL_PTR, &hSession);
	if (rc != CKR_OK) {
		fprintf(stderr, 
			"ERROR call to C_OpenSession failed, rc = 0x%x\n", rc);
		goto out;
	}
	fprintf(stderr, "C_OpenSession success\n");
 
	rc = funcs->C_GetMechanismInfo(SlotId, 
		CKM_GOSTR3411_12_512_HMAC, &minfo);
	if (rc != CKR_OK) {
		fprintf(stderr, 
"\n======== Mechanism CKM_GOSTR3411_12_512_HMAC "
"not supported ===========\n");
	} else {
		fprintf(stderr, 
"\n==================== CKM_GOSTR3411_12_512_HMAC "
"test ==================\n");
		rc = test_gost3411_12_512_hmac(hSession);
		if (rc != CKR_OK) {
			fprintf(stderr, 
				"ERROR CKM_GOSTR3411_12_512_HMAC failed, rc = 0x%x\n",
				rc);
		} else {
			fprintf(stderr, 
				"CKM_GOSTR3411_12_512_HMAC test passed.\n");
		}
	}
 
	if( (ret = funcs->C_CloseSession(hSession)) != CKR_OK ) {
		fprintf(stderr, 
			"Error: C_CloseSession failed with 0x%x\n", ret);
		rc = ret;
	}
	else {
		fprintf(stderr, "C_CloseSession success\n");
	}

out:
	return rc;
}

CK_RV test_gost3411_12_512_hmac(CK_SESSION_HANDLE sess)
{
	int rc = 0;
	CK_BYTE value[512];
	CK_ULONG len;
   	CK_MECHANISM mechanism_desc = 
		{CKM_GOSTR3411_12_512_HMAC, NULL, 0};
   	CK_MECHANISM_PTR mechanism = &mechanism_desc;

	CK_OBJECT_HANDLE keyh = CK_INVALID_HANDLE;
	static CK_OBJECT_CLASS oclass = CKO_SECRET_KEY;
	static CK_KEY_TYPE key_type = CKK_GENERIC_SECRET;
	static CK_BBOOL ltrue = CK_TRUE;

	static CK_BYTE keyval1[] = { 
		0x0b,0x0b,0x0b,0x0b,0x0b,0x0b,0x0b,0x0b, 
		0x0b,0x0b,0x0b,0x0b,0x0b,0x0b,0x0b,0x0b, 
		0x0b,0x0b,0x0b,0x0b,
	};
	static CK_BYTE data1[] = "Hi There";
	static CK_BYTE et1[] = {
		0x86, 0xb6, 0xa0, 0x6b, 0xfa, 0x9f, 0x19, 0x74,
		0xaf, 0xf6, 0xcc, 0xd7, 0xfa, 0x3f, 0x83, 0x5f,
		0x0b, 0xd8, 0x50, 0x39, 0x5d, 0x60, 0x84, 0xef,
		0xc4, 0x7b, 0x9d, 0xda, 0x86, 0x1a, 0x2c, 0xdf,
		0x0d, 0xca, 0xf9, 0x59, 0x16, 0x07, 0x33, 0xd5,
		0x26, 0x9f, 0x65, 0x67, 0x96, 0x6d, 0xd7, 0xa9,
		0xf9, 0x32, 0xa7, 0x7c, 0xd6, 0xf0, 0x80, 0x01,
		0x2c, 0xd4, 0x76, 0xf1, 0xc2, 0xcc, 0x31, 0xbb,
	};

	static CK_BYTE keyval2[] = { 'J','e','f','e' };
	static CK_BYTE data2[] = "what do ya want for nothing?";
	static CK_BYTE et2[] = {
		0x00,0x61,0x75,0x04,0x8e,0x45,0x72,0xac,
		0x61,0x85,0xfa,0xf5,0xff,0xae,0x49,0x62,
		0x97,0xf9,0x99,0x3f,0xf1,0xab,0xb7,0x05,
		0xce,0x43,0xda,0x09,0x51,0x56,0x8d,0x9a,
		0x00,0x61,0x75,0x04,0x8e,0x45,0x72,0xac,
		0x61,0x85,0xfa,0xf5,0xff,0xae,0x49,0x62,
		0x97,0xf9,0x99,0x3f,0xf1,0xab,0xb7,0x05,
		0xce,0x43,0xda,0x09,0x51,0x56,0x8d,0x9a,
	};

	static CK_BYTE keyval3[] = {
		0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,
		0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,
		0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,
		0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,
		0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,
		0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,
		0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,
		0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,
		0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,
		0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,0xaa,
	};
	static CK_BYTE data3[] = 
		"Test Using Larger Than Block-Size Key - Hash Key First";
	static CK_BYTE et3[] = {
		0x91,0x77,0xa0,0x5f,0xee,0xca,0xfa,0x5c,
		0xbb,0x14,0x8b,0x17,0x61,0x63,0x70,0xb5,
		0xd5,0x9b,0x23,0x4c,0x56,0x7e,0x26,0x0d,
		0xda,0xe2,0x67,0x2a,0x5f,0xd3,0x45,0xfa,
		0x91,0x77,0xa0,0x5f,0xee,0xca,0xfa,0x5c,
		0xbb,0x14,0x8b,0x17,0x61,0x63,0x70,0xb5,
		0xd5,0x9b,0x23,0x4c,0x56,0x7e,0x26,0x0d,
		0xda,0xe2,0x67,0x2a,0x5f,0xd3,0x45,0xfa,
	};
	static CK_ATTRIBUTE key_template[] = {
		{ CKA_VALUE, NULL_PTR, 0 },
		{ CKA_CLASS, &oclass, sizeof(oclass) },
		{ CKA_KEY_TYPE, &key_type, sizeof(key_type) },
		{ CKA_SIGN, &ltrue, sizeof(ltrue) },
		{ CKA_VERIFY, &ltrue, sizeof(ltrue) }
	};
	
	static CK_BYTE data4[] = {
	    0x61, 0x62, 0x63, 0x64, 0x62, 0x63, 0x64, 0x65,
	    0x63, 0x64, 0x65, 0x66, 0x64, 0x65, 0x66, 0x67,
	    0x65, 0x66, 0x67, 0x68, 0x66, 0x67, 0x68, 0x69,
	    0x67, 0x68, 0x69, 0x6A, 0x68, 0x69, 0x6A, 0x6B,
	    0x69, 0x6A, 0x6B, 0x6C, 0x6A, 0x6B, 0x6C, 0x6D,
	    0x6B, 0x6C, 0x6D, 0x6E, 0x6C, 0x6D, 0x6E, 0x6F,
	    0x6D, 0x6E, 0x6F, 0x70, 0x6E, 0x6F, 0x70, 0x71,
	    0x0A
	};
	static CK_BYTE keyval_32[] = {
	    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
	    0x38, 0x39, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66,
	    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
	    0x38, 0x39, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66
	};
	// For big-endian byte order
	static CK_BYTE et_32[] = {
	    0x52, 0x67, 0xfb, 0xa5, 0xbf, 0x8c, 0x73, 0xaf,
	    0x26, 0x2c, 0xc2, 0xa0, 0x45, 0x61, 0x86, 0x2e,
	    0x28, 0xce, 0xd8, 0xe9, 0x6f, 0x0a, 0x85, 0xf8,
	    0xf8, 0x8b, 0x59, 0x40, 0xb8, 0xf5, 0x4a, 0x56,
	    0x52, 0x67, 0xfb, 0xa5, 0xbf, 0x8c, 0x73, 0xaf,
	    0x26, 0x2c, 0xc2, 0xa0, 0x45, 0x61, 0x86, 0x2e,
	    0x28, 0xce, 0xd8, 0xe9, 0x6f, 0x0a, 0x85, 0xf8,
	    0xf8, 0x8b, 0x59, 0x40, 0xb8, 0xf5, 0x4a, 0x56
	};
	// For little-endian byte order
	static CK_BYTE et_32_le[] = {
		0x35, 0xe2, 0x38, 0x6f, 0xe1, 0x5d, 0xcf, 0x24,
		0xac, 0x0d, 0xfd, 0x10, 0xfc, 0x8c, 0x12, 0xd4,
		0x5b, 0x0b, 0x45, 0xd4, 0xac, 0x19, 0xe4, 0x17,
		0xa4, 0xf3, 0x99, 0x08, 0xe1, 0x76, 0x5d, 0x54,
		0x63, 0xab, 0x51, 0x6c, 0xaa, 0x65, 0x5e, 0x8a,
		0xae, 0xa0, 0xe9, 0xc8, 0x32, 0x6b, 0x39, 0x8d,
		0x3a, 0x90, 0x3f, 0xb6, 0x06, 0x7e, 0x4f, 0x30,
		0xa9, 0x9c, 0x1d, 0xb3, 0xb7, 0xf2, 0x6e, 0xb1,	
	};
	static CK_BYTE keyval_31[] = {
	    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
	    0x38, 0x39, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66,
	    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
	    0x38, 0x39, 0x61, 0x62, 0x63, 0x64, 0x65
	};
	// For big-endian byte order
	static CK_BYTE et_31[] = {
	    0x98, 0x24, 0x2e, 0xd7, 0x01, 0x14, 0xc6, 0x94,
	    0xbf, 0xfd, 0x10, 0x7b, 0x2f, 0x7e, 0xc3, 0xc7,
	    0x87, 0x14, 0x61, 0xaf, 0x7f, 0x39, 0xf4, 0x05,
	    0xe7, 0x2b, 0xfb, 0x63, 0x36, 0x15, 0xf1, 0xce,
	    0x98, 0x24, 0x2e, 0xd7, 0x01, 0x14, 0xc6, 0x94,
	    0xbf, 0xfd, 0x10, 0x7b, 0x2f, 0x7e, 0xc3, 0xc7,
	    0x87, 0x14, 0x61, 0xaf, 0x7f, 0x39, 0xf4, 0x05,
	    0xe7, 0x2b, 0xfb, 0x63, 0x36, 0x15, 0xf1, 0xce
	};
	// For little-endian byte order
	static CK_BYTE et_31_le[] = {
		0x78, 0xa5, 0xb1, 0xe4, 0xe8, 0x75, 0x59, 0x8b,
		0xc5, 0xf2, 0x2d, 0x4f, 0x92, 0x0b, 0xe3, 0x76,
		0xa9, 0xbc, 0x68, 0xc5, 0x59, 0xfa, 0xbd, 0xb1,
		0x67, 0xcd, 0x78, 0x4e, 0x75, 0x27, 0x62, 0x4d,
		0x59, 0x2d, 0x75, 0x2a, 0x99, 0xfd, 0x06, 0xfd,
		0x1d, 0xc0, 0x66, 0xff, 0x3a, 0xc8, 0x85, 0x3e,
		0x86, 0x71, 0x8b, 0x26, 0xd8, 0x0d, 0x85, 0xd6,
		0xda, 0xb3, 0x12, 0x2c, 0x47, 0xe4, 0x59, 0x0d,	
	};
	static CK_BYTE keyval_33[] = {
	    0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38,
	    0x39, 0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36,
	    0x37, 0x38, 0x39, 0x30, 0x31, 0x32, 0x33, 0x34,
	    0x35, 0x36, 0x37, 0x38, 0x39, 0x30, 0x31, 0x32,
	    0x0A
	};
	static CK_BYTE keyval_33_hash[] = {
	    0x10, 0xff, 0xed, 0x16, 0x36, 0x4d, 0xa5, 0x3a,
	    0x07, 0xc9, 0xba, 0x00, 0x0c, 0xb5, 0x55, 0x31,
	    0xab, 0x53, 0xda, 0xf7, 0x1c, 0x1f, 0xfe, 0x31,
	    0xad, 0x48, 0x81, 0xaf, 0xa1, 0x31, 0xae, 0xbe
	};
	// For big-endian byte order
	static CK_BYTE et_33[] = {
	    0x18, 0x8f, 0x6c, 0xc3, 0x3f, 0x19, 0xbe, 0x1b,
	    0x00, 0xa0, 0x67, 0x80, 0x90, 0xdb, 0xb2, 0x2b,
	    0x5a, 0xaa, 0xb3, 0xec, 0x4e, 0x97, 0x5f, 0x6a,
	    0xa5, 0xe5, 0x2b, 0xf3, 0x39, 0x57, 0x72, 0xeb,
	    0x18, 0x8f, 0x6c, 0xc3, 0x3f, 0x19, 0xbe, 0x1b,
	    0x00, 0xa0, 0x67, 0x80, 0x90, 0xdb, 0xb2, 0x2b,
	    0x5a, 0xaa, 0xb3, 0xec, 0x4e, 0x97, 0x5f, 0x6a,
	    0xa5, 0xe5, 0x2b, 0xf3, 0x39, 0x57, 0x72, 0xeb
	};
	// For little-endian byte order
	static CK_BYTE et_33_le[] = {
		0x25, 0x39, 0x76, 0x8a, 0xa6, 0xa4, 0xae, 0xe2,
		0xa6, 0xb7, 0xb8, 0xb7, 0x9c, 0x85, 0x1f, 0xd1,
		0xce, 0x1a, 0xe7, 0x4b, 0x3c, 0xa6, 0xbc, 0x3d,
		0x6e, 0xeb, 0x5e, 0xef, 0x2b, 0x66, 0x09, 0xa1,
		0x5f, 0x22, 0xb6, 0x85, 0x9f, 0xf7, 0x53, 0x1a,
		0xdf, 0xd0, 0x6b, 0xbd, 0x5a, 0xe1, 0xc4, 0x43,
		0x5c, 0x8f, 0xdc, 0x81, 0x86, 0x0a, 0x03, 0x5a,
		0x52, 0x9c, 0x38, 0xda, 0x22, 0x16, 0xb7, 0x1e,	
	};
	CK_BYTE *state = NULL;
	CK_ULONG state_len = 0;

	key_template[0].pValue = keyval_32; 
	key_template[0].ulValueLen = sizeof(keyval_32);
	rc = funcs->C_CreateObject(sess, 
		key_template, 
		sizeof(key_template)/sizeof(CK_ATTRIBUTE), 
		&keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_CreateObject failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_SignInit(sess, mechanism, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_SignInit failed: 0x%x\n", rc);
		goto end;
	}

	memset(value,0,sizeof(value));
	len = sizeof(value);
	rc = funcs->C_Sign(sess, data4, sizeof(data4), 
		value, &len);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_Sign failed: 0x%x\n", rc);
		goto end;
	}

	if (len != sizeof(et_32_le)) {
		fprintf(stderr, "Invalid length: %d\n", len);
		rc = -2;
		goto end;
	}
	print_hex(value, len);
	if (memcmp(value, et_32_le, len) != 0) {
		fprintf(stderr, "Invalid value\n");
		rc = -3;
		goto end;
	}
	printf("Sign 32_le OK\n");

	rc = funcs->C_VerifyInit(sess, mechanism, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_VerifyInit failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_Verify(sess, data4, sizeof(data4), 
		value, len);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_Verify failed: 0x%x\n", rc);
		goto end;
	}
	printf("Verify 32_le OK\n");

	rc = funcs->C_DestroyObject(sess, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_DestroyObject failed: 0x%x\n", rc);
		goto end;
	}

	key_template[0].pValue = keyval_31; 
	key_template[0].ulValueLen = sizeof(keyval_31);
	rc = funcs->C_CreateObject(sess, 
		key_template, 
		sizeof(key_template)/sizeof(CK_ATTRIBUTE), 
		&keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_CreateObject failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_SignInit(sess, mechanism, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_SignInit failed: 0x%x\n", rc);
		goto end;
	}

	memset(value,0,sizeof(value));
	len = sizeof(value);
	rc = funcs->C_Sign(sess, data4, sizeof(data4), 
		value, &len);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_Sign failed: 0x%x\n", rc);
		goto end;
	}

	if (len != sizeof(et_31_le)) {
		fprintf(stderr, "Invalid length: %d\n", len);
		rc = -4;
		goto end;
	}
	print_hex(value, len);
	if (memcmp(value, et_31_le, len) != 0) {
		fprintf(stderr, "Invalid value\n");
		rc = -5;
		goto end;
	}
	printf("Sign 31_le OK\n");

	rc = funcs->C_VerifyInit(sess, mechanism, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_VerifyInit failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_Verify(sess, data4, sizeof(data4), 
		value, len);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_Verify failed: 0x%x\n", rc);
		goto end;
	}
	printf("Verify 31_le OK\n");

	rc = funcs->C_DestroyObject(sess, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_DestroyObject failed: 0x%x\n", rc);
		goto end;
	}

	key_template[0].pValue = keyval_33; 
	key_template[0].ulValueLen = sizeof(keyval_33);
	rc = funcs->C_CreateObject(sess, 
		key_template, 
		sizeof(key_template)/sizeof(CK_ATTRIBUTE), 
		&keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_CreateObject failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_SignInit(sess, mechanism, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_SignInit failed: 0x%x\n", rc);
		goto end;
	}

	memset(value,0,sizeof(value));
	len = sizeof(value);
	rc = funcs->C_Sign(sess, data4, sizeof(data4), value, &len);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_Sign failed: 0x%x\n", rc);
		goto end;
	}

	if (len != sizeof(et_33_le)) {
		fprintf(stderr, "Invalid length: %d\n", len);
		rc = -8;
		goto end;
	}
	print_hex(value, len);
	if (memcmp(value, et_33_le, len) != 0) {
		fprintf(stderr, "Invalid value\n");
		rc = -9;
		goto end;
	}

	rc = funcs->C_VerifyInit(sess, mechanism, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_VerifyInit failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_Verify(sess, data4, sizeof(data4), value, len);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_Verify failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_DestroyObject(sess, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_DestroyObject failed: 0x%x\n", rc);
		goto end;
	}

	key_template[0].pValue = keyval1; 
	key_template[0].ulValueLen = sizeof(keyval1);
	rc = funcs->C_CreateObject(sess, 
		key_template, sizeof(key_template)/sizeof(CK_ATTRIBUTE),
		&keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_CreateObject failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_SignInit(sess, mechanism, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_SignInit failed: 0x%x\n", rc);
		goto end;
	}

	memset(value,0,sizeof(value));
	len = sizeof(value);
	rc = funcs->C_Sign(sess, data1, strlen(data1), value, &len);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_Sign failed: 0x%x\n", rc);
		goto end;
	}

	if (len != sizeof(et1)) {
		fprintf(stderr, "Invalid length: %d\n", len);
		rc = -10;
		goto end;
	}
	print_hex(value, len);
	if (memcmp(value, et1, len) != 0) {
		fprintf(stderr, "Invalid value\n");
		rc = -11;
		goto end;
	}

	rc = funcs->C_VerifyInit(sess, mechanism, keyh);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_VerifyInit failed: 0x%x\n", rc);
		goto end;
	}

	rc = funcs->C_Verify(sess, data1, strlen(data1), value, len);
	if (rc != CKR_OK) {
		fprintf(stderr, "C_Verify failed: 0x%x\n", rc);
		goto end;
	}
	printf("SUCCESS\n");
	rc = CKR_OK;
end:
	if (keyh != CK_INVALID_HANDLE) {
		funcs->C_DestroyObject(sess, keyh);
	}

	return rc;
}

